/* Struct to connect to ebshared global kernel variable */
/* Shares data between device drivers ... */
/* fields are defined as ints here, but will be structs in drivers */

#ifndef EBhInc
#define EBhInc
#include <time.h>

#include <ebstats.h>
#include <rawdata.h>
#include <shmhist.h>
#include <cluster.h>
#include <clusterdata.h>
#include <clusteroutput.h>
#include <ebspec.h>
#include <rundata.h>
#include <semaphore.h>

#include <eventsearch.h>

/* Data transfer buffer queue  definitions ( must stay aligned) ... */

#define MAX_RAWDATABUFFERS          1024
#define MAX_PREFILTERBUFFERS         256
#define MAX_OUTPUTBUFFERS            256

#define MAX_RAWDATABUFFERSIZE       65536
#define MAX_PREFILTERBUFFERSIZE     MAX_RAWDATABUFFERSIZE
#define MAX_OUTPUTBUFFERSIZE        MAX_RAWDATABUFFERSIZE
 
struct data_transfer {
  volatile int start;
  volatile int end;
  volatile int length;
  volatile int maxlength;
  };

/* Cluster buffering ... */

#define MAX_CLUSTERS              32

#define MAX_CLUSTERBUFFERS     32768
#define MAX_CLUSTERBUFFERSIZE   8192

#define MAXBUFFERSPERCLUSTER     256

#define MAX_CLUSTERQSLOTS       MAX_CLUSTERBUFFERS

/*
Slot_firstentry is the first buffer record number in the current slot position.
It is the absolute buffer record position starting from zero.
Slot_nentries is the number of buffer records in the current slot position
*/

typedef struct clusterqueue {
  volatile TimeStamp64 slot_endTS[MAX_CLUSTERQSLOTS+1];
  volatile short slot_nentries[MAX_CLUSTERQSLOTS+1];
  volatile void *slot_buffers[MAX_CLUSTERQSLOTS+1];
} ClusterQ;

/* Data I/O descriptor */

typedef struct io_channel {
  int enabled;
  int handler;
  int stream;
  int status;
  char device[16];
  char volname[8];
  char filename[128];
} IOchannel;

/* Shared memory layout ... */

#ifdef Linux
#ifdef SIXTYFOURBIT
#define SHM_ATTACH_BASE       0x1100000000ULL
#define SYSTEMSEGMENTSBASE    0x1180000000ULL
#define RAWDATASEGMENTSBASE   0x1190000000ULL
#define PREFILTERSEGMENTSBASE 0x11a0000000ULL
#define CLUSTERSEGMENTSBASE   0x11b0000000ULL
#define OUTPUTSEGMENTSBASE    0x11c0000000ULL
#else
#define SHM_ATTACH_BASE       0x60000000
#define SYSTEMSEGMENTSBASE    0x68000000
#define RAWDATASEGMENTSBASE   0x64000000
#define PREFILTERSEGMENTSBASE 0x6c000000
#define CLUSTERSEGMENTSBASE   0x70000000
#define OUTPUTSEGMENTSBASE    0x6e000000
#endif
#endif

#define SYSTEM_KEYTYPE 0
#define RAWDATABUFFERS 1
#define CLUSTERBUFFERS 2
#define OUTPUTBUFFERS  3
#define PREFILTERBUFFERS 4

#define GREAT_EB_SHM_BASE            0x19070000
#define GREAT_EB_SHM_KEY             GREAT_EB_SHM_BASE

#define MAX_OUTPUT_CHANNELS 4
#define MAX_OUTPUT_STREAMS MAX_OUTPUT_CHANNELS

/* Note: event output modes must start at ONE */

enum { NULL_CHANNEL,
       RAWDATA_CHANNEL,
       FILTERED_CHANNEL,
       EVENT1_CHANNEL,
       EVENT2_CHANNEL,
       CLUSTERS_CHANNEL,
       MAX_OUTPUT_MODES
};

enum { RAWDATA_QUEUE,
       FILTERED_QUEUE,
       CLUSTERS_QUEUE,
       EVENTS_QUEUE,
       MAX_OUTPUT_QUEUES
};

enum { PREFILTER_OPTION0, PREFILTER_OPTION1, PREFILTER_OPTION2 };

struct ebshared {
  int trace;
  int debug;
  int version;
  int new_state;             /* Command to execute state change */

  int systemState;
  int inputState;
  int prefilterState;
  int clusterState;

  int eventState;
  int outputState;
  int transmit_state_event;
  int transmit_state_output;

  int transmit_state_cluster;
  int prefilter_enable;
  int enable_clusteroutput;    /* Output cluster data to tape */
  int enable_eventsearch;      /* Allow eventsearching */

  int outputblocksize;
  int zero_on_go;
  int stop_on_flag;
  int input_reject_mode;

  Stat *statistics;
  struct statistics_record stats;

  int scope_mode;
  int trigger_label;
  int trigger_offset;
  int trigger_padding2;
  TimeStamp64 scope_triggerTS;

  void *rawdatabufferspace;
  int maxrawdatabuffers;
  int maxrawdatabuffersize;
  int rawdatabufferoffset;

  void *prefilterbufferspace;
  int maxprefilterbuffers;
  int maxprefilterbuffersize;
  int prefilterbufferoffset;

  void *clusterbufferspace;
  int maxclusterbuffers;
  int maxclusterbuffersize;
  int maxclusters;

  void *outputbufferspace;
  int maxoutputbuffers;
  int maxoutputbuffersize;
  int padding6;

  /* RawData buffers ... */

  void *rawdatabuffers[MAX_RAWDATABUFFERS];    /* Original/Total set */

  volatile struct data_transfer rawdatafree;
  void *rawdatafreebuffers[MAX_RAWDATABUFFERS+1];

  volatile struct data_transfer rawdataready;
  void *rawdatareadybuffers[MAX_RAWDATABUFFERS+1];
  int rawdatalength[MAX_RAWDATABUFFERS+1];

  /* PreFilter buffers ... */

  void *prefilterbuffers[MAX_PREFILTERBUFFERS];    /* Original/Total set */

  volatile struct data_transfer prefilter1free;
  void *prefilter1freebuffers[MAX_PREFILTERBUFFERS+1];
  volatile struct data_transfer prefilter2free;
  void *prefilter2freebuffers[MAX_PREFILTERBUFFERS+1];

  volatile struct data_transfer prefilter1ready;
  void *prefilter1readybuffers[MAX_PREFILTERBUFFERS+1];
  int prefilter1length[MAX_PREFILTERBUFFERS+1];
  volatile struct data_transfer prefilter2ready;
  void *prefilter2readybuffers[MAX_PREFILTERBUFFERS+1];
  int prefilter2length[MAX_PREFILTERBUFFERS+1];

  volatile struct data_transfer prefilterfree;
  void *prefilterfreebuffers[MAX_PREFILTERBUFFERS+1];
  volatile struct data_transfer prefilterready;
  void *prefilterreadybuffers[MAX_PREFILTERBUFFERS+1];
  int prefilterlength[MAX_PREFILTERBUFFERS+1];

  /* Cluster to EventSearch buffers ... */

  void *clusterbuffers[MAX_CLUSTERBUFFERS];   /* Original/Total set */

  volatile struct data_transfer clusterfree;
  void *clusterfreebuffers[MAX_CLUSTERBUFFERS+1];

  volatile struct data_transfer clusterready;
  volatile int ClusterQ_slot_cluster[MAX_CLUSTERQSLOTS+1];
  volatile void *ClusterQ_slot_buffer[MAX_CLUSTERQSLOTS+1];
  volatile TimeStamp64 ClusterQ_slot_endTS[MAX_CLUSTERQSLOTS+1];

  /* EventSearch input cluster buffers ... */

  ClusterQ ClusterQ[MAX_CLUSTERS];
  volatile int ClusterQ_slot_start[MAX_CLUSTERS];
  volatile int ClusterQ_slot_current[MAX_CLUSTERS];
  volatile int ClusterQ_slot_end[MAX_CLUSTERS];

  /* Output buffers ... */

  void *outputbuffers[MAX_OUTPUTBUFFERS];     /* Original/Total set */

  volatile struct data_transfer outputfree[MAX_OUTPUT_QUEUES];
  void *outputfreebuffers[MAX_OUTPUT_QUEUES][MAX_OUTPUTBUFFERS+1];

  volatile struct data_transfer outputready[MAX_OUTPUT_QUEUES];
  void *outputreadybuffers[MAX_OUTPUT_QUEUES][MAX_OUTPUTBUFFERS+1];
  int outputlength[MAX_OUTPUT_QUEUES][MAX_OUTPUTBUFFERS+1];
  
/* Prefilter definitions filled out from setup ... */

  int fast_slow[MAXLABELS+1];
  int fast_slow_delay;
  int fast_slow_width;

  int bgo_filter_bgoparam[MAXLABELS+1];
  int bgo_filter_geparam[MAXLABELS+1][MAX_PATTERN_BITS];
  int bgo_filter_width;
  int bgo_filter_mode;
  int bgo_filter_pileup_mode;
  /* fold1 applies to bgofilter and fastfilter outside of window */
  /* fold2 applies to within fast filter window iff fast filter enabled */
  int bgo_filter_fold1width;
  int bgo_filter_fold1mode;
  int bgo_filter_fold2width;
  int bgo_filter_fold2mode;

/* Cluster definitions filled out from setup ... */

  int nclusters;
  char cluster_type_name[MAX_CLUSTERS][NAMELEN];
  char cluster_name_name[MAX_CLUSTERS][NAMELEN];
  ClusterDesc clusterdesc[MAX_CLUSTERS];
  
  int clusterfromparam[MAXLABELS+1];
  int clusterinstancefromparam[MAXLABELS+1];
  int clusteroffsetfromparam[MAXLABELS+1];
  int param_type[MAXLABELS+1];
  int tsadjust[MAXLABELS];

  int paramGainEnable[MAXLABELS+1];
  float paramGainA[MAXLABELS+1];
  float paramGainB[MAXLABELS+1];
  float paramGainC[MAXLABELS+1];

/* EventSearch definitions filled out from setup ... */

  int nnodes;
  EventSearchDesc eventsearchdesc[MAX_NODES];

  /* Data output channel definitions ... */

  IOchannel InChannel;
  IOchannel OutChannel[MAX_OUTPUT_CHANNELS];

  RunInfo InRunInfo[MAX_RUNINFO_RECORDS];
  int NInRunInfoRecords;
  int NewInRunInfoRecords;

  RunInfo OutRunInfo[MAX_OUTPUT_CHANNELS][MAX_RUNINFO_RECORDS];
  int NOutRunInfoRecords[MAX_OUTPUT_CHANNELS];
  int NewOutRunInfoRecords[MAX_OUTPUT_CHANNELS];
  int CurrentOutRunInfoRecord[MAX_OUTPUT_CHANNELS];

  /* Internal data mode outputs ... */

  int OutModeEnable[MAX_OUTPUT_MODES];
  TimeStamp64 *outdata[MAX_OUTPUT_MODES];
  TimeStamp64 *outdata_start[MAX_OUTPUT_MODES];
  TimeStamp64 *outdata_bufend[MAX_OUTPUT_MODES];

  /* Shmid of basic shared memory table */

  int sysShmid;    
  int spectrum_space;
  int nspectra;
  struct spectable spectable[MAXSPECTRA];
  char experiment[256];
  char specdir[256];
  char basedir[256];
  char bindir[256];
  char libdir[256];
  char configdir[256];
  char setupdir[256];
  char cluster_setup_filename[256];
  char eventsearch_setup_filename[256];

  // Semaphores ...

  sem_t bufferaccess_sem_key;
  sem_t rawdatabufferfree_sem_key;
  sem_t rawdatabufferready_sem_key;
  sem_t prefilterbufferfree_sem_key;
  sem_t prefilterbufferready_sem_key;
  sem_t prefilter1bufferfree_sem_key;
  sem_t prefilter1bufferready_sem_key;
  sem_t prefilter2bufferfree_sem_key;
  sem_t prefilter2bufferready_sem_key;
  sem_t clusterbufferfree_sem_key;
  sem_t clusterbufferready_sem_key;
  sem_t outputbufferfree_sem_key[MAX_OUTPUT_CHANNELS];
  sem_t outputbufferready_sem_key;

};

typedef struct ebshared EBshared;

/* Data transfer buffer status conditions ... */

#define BUFFER_FREE                             0
#define BUFFER_IN_USE                           1
#define BUFFER_REMOTE                           2
#define BUFFER_READY                            3

/* Data transfer buffer control record formats ... */
/* "address" is hsm module address or kernel virtual for epq within ppc */
/* "physaddr" is vme addr for hsm, local phys addr for dma usage */

struct eg_data_struct {
  int status;
  int *address;
  int length;
  short sequence;
  short stream;
  int procnum;
  int *physaddr;
  int spare1;
  int boardaddr;
  };

/* IO communication command types ... */

#define INPUT_ACK                               0
#define INPUT_REQUEST                           1
#define OUTPUT_REQUEST                          2

/* Command decoding */

enum {NOACTION=0, INITIALISE, GO, STOP, FLUSH, EXIT, N_COMMANDS};

/* State decoding */

enum {UNINITIALISED=0, 
      INITIALISEREADY,
      POST_INITIALISEREADY,
      INITIALISED,
      GOINGREADY,
      GOING,
      FLUSHING,
      FLUSHED,
      STOPPING,
      STOPPED,
      EXITING,
      N_STATES};


#include <ebdefs.h>
#include <msg.h>

#include <ebdebug.h>

#endif

#ifndef _shmC
extern void *create_shared_memory(char *, void *, int size);
#endif

#ifndef _bufsC
extern void *cluster_free_buffer_get();
extern void *cluster_ready_buffer_get(int *cluster, TimeStamp64 *endTS);
extern void *rawdata_free_buffer_get();
extern void *rawdata_ready_buffer_get(int *len);
extern void *prefilter1_free_buffer_get();
extern void *prefilter2_free_buffer_get();
extern void *prefilter_free_buffer_get();
extern void *prefilter1_ready_buffer_get(int *len);
extern void *prefilter2_ready_buffer_get(int *len);
extern void *prefilter_ready_buffer_get(int *len);
extern void *output_free_buffer_get(int mode);
extern void *output_ready_buffer_get(int *datalen, int *mode);
#endif

/* ======================================================================= */



